##
#     Project: gWakeOnLAN
# Description: Wake up your machines using Wake on LAN
#      Author: Fabio Castelli (Muflone) <muflone@muflone.com>
#   Copyright: 2009-2022 Fabio Castelli
#     License: GPL-3+
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
##

import logging
import os.path
import socket

from gwakeonlan.constants import FILE_ARP_CACHE
from gwakeonlan.models.abstract import ModelAbstract
from gwakeonlan.models.arpcache_item import ArpCacheItem


class ModelArpCache(ModelAbstract):
    COL_IPADDRESS = 0
    COL_MACADDRESS = 1
    COL_HOSTNAME = 2

    def add_data(self, item):
        """Add a new row to the model if it doesn't exist"""
        super(self.__class__, self).add_data(item)
        if item.ip_address not in self.rows:
            new_row = self.model.append((
                item.ip_address,
                item.mac_address.upper(),
                item.hostname
            ))
            self.rows[item.ip_address] = new_row

    def refresh(self):
        """Clear the model and reload all the hosts from the ARP cache file"""
        self.clear()
        # Read ARP cache file
        if os.path.isfile(FILE_ARP_CACHE):
            try:
                arpf = open(FILE_ARP_CACHE, 'r')
                # Skip first and last line
                for line in arpf.readlines()[1:]:
                    if line:
                        # Add IP Address and MAC address to the model
                        logging.debug(f'arp line:\n{line}')
                        arp_ip = line[:17].rstrip()
                        arp_mac = line[41:58].upper()
                        # Skip incomplete MAC addresses
                        if arp_mac != '00:00:00:00:00:00':
                            detected_hostname = socket.getfqdn(arp_ip)
                            # I will not trust of getfqdn if the returned
                            # hostname is the same of the source IP address
                            if detected_hostname == arp_ip:
                                detected_hostname = ''
                            logging.info(f'discovered {arp_ip} '
                                         f'with address {arp_mac}')
                            self.add_data(ArpCacheItem(
                                ip_address=arp_ip,
                                mac_address=arp_mac,
                                hostname=detected_hostname
                            ))
                arpf.close()
            except (FileNotFoundError, PermissionError):
                logging.error(f'unable to read {FILE_ARP_CACHE}')

    def get_ip_address(self, treeiter):
        """Returns the IP address for the selected TreeIter"""
        return self.model[treeiter][self.COL_IPADDRESS]

    def get_mac_address(self, treeiter):
        """Returns the MAC address for the selected TreeIter"""
        return self.model[treeiter][self.COL_MACADDRESS]

    def get_hostname(self, treeiter):
        """Returns the hostname for the selected TreeIter"""
        return self.model[treeiter][self.COL_HOSTNAME]



##
#     Project: gWakeOnLAN
# Description: Wake up your machines using Wake on LAN
#      Author: Robin Schneider <ypid@riseup.net>
#   Copyright: 2015 Robin Schneider <ypid@riseup.net>
#     License: GPL-3+
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
##

from gwakeonlan.constants import BROADCAST_ADDRESS, DEFAULT_UDP_PORT
from gwakeonlan.models.machine_item import MachineItem


class ImportEthers(object):
    def __init__(self, import_l3_dest=BROADCAST_ADDRESS):
        self.import_l3_dest = import_l3_dest

    def import_file(self, filepath, model, icon):
        with open(filepath, 'r') as import_fh:
            for line in import_fh:
                line = line.split(sep='#', maxsplit=1)[0].strip()
                if line:
                    mac_address, machine_name = line.split()
                    model.add_data(MachineItem(
                        name=machine_name,
                        mac_address=(mac_address
                                     .replace('.', ':')
                                     .replace(' ', ':')
                                     .replace('-', ':')),
                        port_number=DEFAULT_UDP_PORT,
                        destination=self.import_l3_dest,
                        icon=icon))



##
#     Project: gWakeOnLAN
# Description: Wake up your machines using Wake on LAN
#      Author: Fabio Castelli (Muflone) <muflone@muflone.com>
#   Copyright: 2009-2022 Fabio Castelli
#     License: GPL-3+
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
##

from gwakeonlan.constants import BROADCAST_ADDRESS
from gwakeonlan.models.abstract import ModelAbstract


class ModelMachines(ModelAbstract):
    COL_SELECTED = 1
    COL_MACADDRESS = 2
    COL_REQUESTTYPE = 3
    COL_DESTINATION = 4
    COL_PORTNR = 5
    COL_ICON = 6

    def add_data(self, item):
        """Add a new row to the model if it doesn't exist"""
        super(self.__class__, self).add_data(item)
        if item.name not in self.rows:
            new_row = self.model.append((
                item.name,
                False,
                item.mac_address.upper(),
                'Local'
                if item.destination == BROADCAST_ADDRESS
                else 'Internet',
                item.destination,
                item.port_number,
                item.icon
            ))
            self.rows[item.name] = new_row

    def get_selected(self, treeiter):
        """Return if the TreeIter is selected"""
        return self.model[treeiter][self.COL_SELECTED]

    def set_selected(self, treeiter, value):
        """Set the TreeIter selection"""
        self.model[treeiter][self.COL_SELECTED] = value

    def get_machine_name(self, treeiter):
        """Return the machine name from a TreeIter"""
        return self.model[treeiter][self.COL_KEY]

    def set_machine_name(self, treeiter, value):
        """Set the machine name for a TreeIter"""
        self.model[treeiter][self.COL_KEY] = value

    def get_mac_address(self, treeiter):
        """Return the MAC address from a TreeIter"""
        return self.model[treeiter][self.COL_MACADDRESS]

    def set_mac_address(self, treeiter, value):
        """Set the MAC address for a TreeIter"""
        self.model[treeiter][self.COL_MACADDRESS] = value

    def get_request_type(self, treeiter):
        """Return the request type from a TreeIter"""
        return self.model[treeiter][self.COL_REQUESTTYPE]

    def get_destination(self, treeiter):
        """Return the destination from a TreeIter"""
        return self.model[treeiter][self.COL_DESTINATION]

    def set_destination(self, treeiter, value):
        """Set the destination for a TreeIter"""
        self.model[treeiter][self.COL_DESTINATION] = value
        self.model[treeiter][self.COL_REQUESTTYPE] = (
            'Local' if value == BROADCAST_ADDRESS else 'Internet')

    def get_port_number(self, treeiter):
        """Return the port number from a TreeIter"""
        return self.model[treeiter][self.COL_PORTNR]

    def set_port_number(self, treeiter, value):
        """Set the port number for a TreeIter"""
        self.model[treeiter][self.COL_PORTNR] = value

    def set_icon(self, treeiter, value):
        """Set the background color for a TreeIter"""
        self.model[treeiter][self.COL_ICON] = value



##
#     Project: gWakeOnLAN
# Description: Wake up your machines using Wake on LAN
#      Author: Fabio Castelli (Muflone) <muflone@muflone.com>
#   Copyright: 2009-2022 Fabio Castelli
#     License: GPL-3+
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
##

import itertools
import logging
import pathlib
import struct
import socket

from gi.repository import Gtk
from gi.repository import GdkPixbuf

from gwakeonlan.constants import DIR_UI


def format_mac_address(mac):
    """Return the mac address formatted with colon"""
    mac = mac.replace(':', '').replace('.', '').replace('-', '')
    return ':'.join([mac[i:i+2] for i in range(0, len(mac), 2)]).upper()


def get_pixbuf_from_icon_name(icon_name, size):
    """Get a Gdk.PixBuf from a theme icon"""
    theme = Gtk.IconTheme.get_default()
    path_icon_name = pathlib.Path(icon_name)
    if theme.has_icon(icon_name):
        # The icon was a theme icon
        icon = theme.load_icon(icon_name=icon_name,
                               size=size,
                               flags=Gtk.IconLookupFlags.USE_BUILTIN)
    elif theme.has_icon(path_icon_name.stem):
        # The theme contains an icon with the same file name
        icon = theme.load_icon(icon_name=path_icon_name.stem,
                               size=size,
                               flags=Gtk.IconLookupFlags.USE_BUILTIN)
    elif path_icon_name.is_file():
        # The icon was a full filename
        icon = GdkPixbuf.Pixbuf.new_from_file(icon_name)
    else:
        # The icon was not found in the current theme, search for filenames
        # with png or jpg extensions
        if path_icon_name.suffix.lower() in ('.png', '.jpg', '.xpm', '.svg'):
            filenames = (icon_name, )
        else:
            filenames = (f'{icon_name}.png',
                         f'{icon_name}.jpg',
                         f'{icon_name}.xpm',
                         f'{icon_name}.svg')
        # Search for filenames in icons and pixmaps directories
        icon = None
        search_in_paths = ('/usr/share/icons',
                           '/usr/share/pixmaps')
        for path, filename in itertools.product(search_in_paths, filenames):
            file_path = pathlib.Path(path) / filename
            if file_path.is_file():
                icon = GdkPixbuf.Pixbuf.new_from_file(str(file_path))
                break
    if icon:
        # If size is not correct then resize the icon to the requested size
        if icon.get_width() != size or icon.get_height() != size:
            icon = icon.scale_simple(size, size,
                                     GdkPixbuf.InterpType.BILINEAR)
    else:
        logging.warning(f'missing icon: {icon_name}')
    return icon


def get_treeview_selected_row(widget):
    """Return the selected row in a GtkTreeView"""
    return widget.get_selection().get_selected()[1]


def get_ui_file(filename):
    """Return the full path of a Glade/UI file"""
    return str(DIR_UI / filename)


def process_events():
    """Process every pending GTK+ event"""
    while Gtk.events_pending():
        Gtk.main_iteration()


def readlines(filename, empty_lines=False):
    """Read all the lines of a filename, optionally skipping empty lines"""
    result = []
    with open(filename) as f:
        for line in f.readlines():
            line = line.strip()
            if line or empty_lines:
                result.append(line)
        f.close()
    return result


def show_message_dialog_yesno(parent, message, title, default_response):
    """Show a GtkMessageDialog with yes and no buttons"""
    dialog = Gtk.MessageDialog(
        parent=parent,
        flags=Gtk.DialogFlags.MODAL,
        type=Gtk.MessageType.QUESTION,
        buttons=Gtk.ButtonsType.YES_NO,
        message_format=message
    )
    dialog.set_title(title)
    if default_response:
        dialog.set_default_response(default_response)
    response = dialog.run()
    dialog.destroy()
    return response


def wake_on_lan(mac_address, port_number, destination):
    """Turn on remote machine using Wake On LAN"""
    logging.info(f'turning on: {mac_address} '
                 f'through {destination} '
                 f'using port number {port_number}')
    # Magic packet (6 times FF + 16 times MAC address)
    packet = 'FF' * 6 + mac_address.replace(':', '') * 16
    data = []
    for i in range(0, len(packet), 2):
        data.append(struct.pack('B', int(packet[i:i+2], 16)))

    # Send magic packet to the destination
    logging.info(f'sending packet {packet} [{len(packet)}/{len(data)}]\n')
    sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    sock.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
    if destination == '255.255.255.255':
        destination = '<broadcast>'
    data = b''.join(data)
    for _ in range(10):
        sock.sendto(data, (destination, port_number))



#!/usr/bin/env python3
##
#     Project: gWakeOnLAN
# Description: Wake up your machines using Wake on LAN
#      Author: Fabio Castelli (Muflone) <muflone@muflone.com>
#   Copyright: 2009-2022 Fabio Castelli
#     License: GPL-3+
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
##

import itertools
import pathlib
import setuptools
import setuptools.command.install_scripts
import subprocess

# Importing distutils after setuptools uses the setuptools distutils
from distutils.command.install_data import install_data

from gwakeonlan.constants import (APP_AUTHOR,
                                  APP_AUTHOR_EMAIL,
                                  APP_DESCRIPTION,
                                  APP_DOMAIN,
                                  APP_NAME,
                                  APP_VERSION,
                                  URL_APPLICATION,
                                  URL_SOURCES)


class InstallScripts(setuptools.command.install_scripts.install_scripts):
    def run(self):
        setuptools.command.install_scripts.install_scripts.run(self)
        self.rename_python_scripts()

    def rename_python_scripts(self):
        """Rename main executable python script without .py extension"""
        for script in self.get_outputs():
            path_file_script = pathlib.Path(script)
            path_destination = path_file_script.with_suffix(suffix='')
            if path_file_script.suffix == '.py':
                # noinspection PyUnresolvedReferences
                setuptools.distutils.log.info(
                    'renaming the python script '
                    f'{path_file_script.name} -> '
                    f'{path_destination.stem}')
                path_file_script.rename(path_destination)


class InstallData(install_data):
    def run(self):
        self.install_icons()
        self.install_translations()
        install_data.run(self)

    def install_icons(self):
        # noinspection PyUnresolvedReferences
        setuptools.distutils.log.info('Installing icons...')
        path_icons = pathlib.Path('share') / 'icons' / 'hicolor'
        for path_format in pathlib.Path('icons').iterdir():
            self.data_files.append((path_icons / path_format.name / 'apps',
                                    list(map(str, path_format.glob('*')))))

    def install_translations(self):
        # noinspection PyUnresolvedReferences
        setuptools.distutils.log.info('Installing translations...')
        path_base = pathlib.Path(__file__).parent.absolute()
        # Find where to save the compiled translations
        try:
            # Use the install_data (when using "setup.py install --user")
            # noinspection PyUnresolvedReferences
            path_install = pathlib.Path(self.install_data)
        except AttributeError:
            # Use the install_dir (when using "setup.py install")
            path_install = pathlib.Path(self.install_dir)
        path_locale = path_install / 'share' / 'locale'
        for path_file_po in pathlib.Path('po').glob('*.po'):
            path_destination = path_locale / path_file_po.stem / 'LC_MESSAGES'
            path_file_mo = path_destination / f'{APP_DOMAIN}.mo'

            if not path_destination.exists():
                # noinspection PyUnresolvedReferences
                setuptools.distutils.log.info(f'creating {path_destination}')
                path_destination.mkdir(parents=True)

            # noinspection PyUnresolvedReferences
            setuptools.distutils.log.info(f'compiling {path_file_po} -> '
                                          f'{path_file_mo}')
            subprocess.call(
                args=('msgfmt',
                      f'--output-file={path_file_mo}',
                      path_file_po),
                cwd=path_base)


class CommandCreatePOT(setuptools.Command):
    description = "create base POT file"
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        path_base = pathlib.Path(__file__).parent.absolute()
        path_po = path_base / 'po'
        path_ui = path_base / 'ui'
        path_pot = path_po / f'{APP_DOMAIN}.pot'
        list_files_process = []
        # Add *.ui files to list of files to process
        for filename in path_ui.glob('*.ui'):
            list_files_process.append(filename.relative_to(path_base))
        # Add *.py files to list of files to process
        for filename in path_base.rglob('*.py'):
            list_files_process.append(filename.relative_to(path_base))
        # Sort the files to process them always in the same order (hopefully)
        list_files_process.sort()
        # Extract messages from the files to process
        # noinspection PyTypeChecker
        subprocess.call(
            args=itertools.chain((
                'xgettext',
                '--keyword=_',
                '--keyword=N_',
                f'--output={path_pot}',
                '--add-location',
                f'--package-name={APP_NAME}',
                f'--copyright-holder={APP_AUTHOR}',
                f'--msgid-bugs-address={URL_SOURCES}issues/'),
                list_files_process),
            cwd=path_base)


class CommandCreatePO(setuptools.Command):
    description = "create translation PO file"
    user_options = [
        ('locale=', None, 'Define locale'),
        ('output=', None, 'Define output file'),
        ]

    # noinspection PyUnusedLocal
    def __init__(self, dist, **kw):
        super().__init__(dist)
        self.locale = None
        self.output = None

    def initialize_options(self):
        pass

    def finalize_options(self):
        assert self.locale, 'Missing locale'
        assert self.output, 'Missing output file'

    def run(self):
        path_base = pathlib.Path(__file__).parent.absolute()
        path_file_pot = path_base / 'po' / f'{APP_DOMAIN}.pot'
        path_file_po = path_base / 'po' / f'{self.output}.po'
        # Create PO file
        subprocess.call(
            args=('msginit',
                  f'--input={path_file_pot}',
                  '--no-translator',
                  f'--output-file={path_file_po}',
                  f'--locale={self.locale}'),
            cwd=path_base)


class CommandTranslations(setuptools.Command):
    description = "build translations"
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        path_base = pathlib.Path(__file__).parent.absolute()
        path_po = path_base / 'po'
        for file_po in path_po.glob('*.po'):
            subprocess.call(('msgmerge',
                             '--update',
                             '--backup=off',
                             file_po,
                             path_po / f'{APP_DOMAIN}.pot'))
            path_mo = path_base / 'locale' / file_po.stem / 'LC_MESSAGES'
            if not path_mo.exists():
                path_mo.mkdir(parents=True)
            file_mo = path_mo / f'{APP_DOMAIN}.mo'
            subprocess.call(('msgfmt',
                             '--output-file',
                             file_mo,
                             file_po))


setuptools.setup(
    name=APP_NAME,
    version=APP_VERSION,
    author=APP_AUTHOR,
    author_email=APP_AUTHOR_EMAIL,
    maintainer=APP_AUTHOR,
    maintainer_email=APP_AUTHOR_EMAIL,
    url=URL_APPLICATION,
    description=APP_DESCRIPTION,
    license='GPL v3',
    scripts=['gwakeonlan.py'],
    packages=['gwakeonlan',
              'gwakeonlan.models',
              'gwakeonlan.ui'],
    data_files=[
        (f'share/{APP_DOMAIN}/data',
            ['data/gwakeonlan.png']),
        (f'share/{APP_DOMAIN}/data/icons',
            list(map(str, (pathlib.Path('data') / 'icons').glob('*')))),
        ('share/applications',
            ['data/gwakeonlan.desktop']),
        (f'share/doc/{APP_DOMAIN}',
         list(itertools.chain(
             list(map(str, pathlib.Path('doc').glob('*'))),
             list(map(str, pathlib.Path('.').glob('*.md')))))),
        ('share/man/man1',
            ['man/gwakeonlan.1']),
        (f'share/{APP_DOMAIN}/ui', [str(file)
                                    for file
                                    in pathlib.Path('ui').glob('*')
                                    if not file.name.endswith('~')]),
        ('share/metainfo',
         ['data/com.muflone.gwakeonlan.metainfo.xml']),
    ],
    cmdclass={
        'install_scripts': InstallScripts,
        'install_data': InstallData,
        'create_pot': CommandCreatePOT,
        'create_po': CommandCreatePO,
        'translations': CommandTranslations
    }
)
